@@ -17,7 +17,7 @@
 
 ;;; Notes:
 
-;; TODO: be more agressive to parse 123 as fixed and 567.0 as float.
+;; TODO: be more aggressive to parse 123 as fixed and 567.0 as float.
 ;; We can convert to string later.
 
 ;; args string => (arg-list (arg "abc") (opt-arg "def" "123") (rest "args"))
@@ -44,8 +44,7 @@
   #:use-module (nyacc lang sx-util)
   #:use-module (sxml match)
   #:use-module ((srfi srfi-1) #:select (fold-right))
-  #:use-module (ice-9 match)
-  )
+  #:use-module (ice-9 match))
 
 (use-modules (ice-9 pretty-print))
 (define pp pretty-print)
@@ -108,10 +107,18 @@
 
 (define noop-command '(command (string "NOOP")))
 
+;; A command is a sequence of words.
+;; A word is a sequence of fragments.
+
+;; BUG in read-frag? if ending found two levels deep need to pop two levels ??
+;; in read-frag: maybe tagged as string when should not be
+
 (define (read-command port)
   ;; This is a bit of a hack job.
   (letrec
-      ((error
+      ((srcf (port-filename port))
+
+       (error
 	(lambda (fmt . args)
 	  (with-input-from-port port
 	    (lambda () (apply report-error port fmt args)))))
@@ -178,9 +185,7 @@
 	      (read-char port) ;; {
 	      (loop (cons (read-brace) wordl) (peek-char port)))
 
-	     ;;((eq? end-cs cs:nl)
 	     ((eq? end-cs cs:ct)
-	      ;;(let ((word (read-word cs:nl+ws)))
 	      (let ((word (read-word cs:ct+ws)))
 		(loop (foldin word wordl) (peek-char port))))
 	     
@@ -197,7 +202,10 @@
 	    (cond
 	     ((eq? frag end-cs)
 	      (db "W:   done\n")
-	      (cons 'word (reverse fragl)))
+              (let* ((tail (reverse fragl)) (word (cons 'word tail)))
+                (when (pair? tail)
+                  (set-source-properties! word (source-properties (car tail))))
+                word))
 	     (else
 	      (loop (cons frag fragl) (read-frag end-cs)))))))
 
@@ -205,13 +213,13 @@
 	(lambda (tag chl ch)
 	  (db "F:     finish ~S ~S ~S\n" tag chl ch)
 	  (unless (eof-object? ch) (unread-char ch port))
-	  ;;(case tag ((string) (rls chl)) (else (list tag (rls chl))))))
 	  (list tag (rls chl))))
 
        (swallow
 	(lambda (val)
 	  (db "swallow ")
-	  (read-char port) val))
+	  (read-char port)
+          val))
 
        (read-brace
 	(lambda ()
@@ -230,10 +238,9 @@
        
        (read-index
 	(lambda ()
-	  ;;(if (not (char=? #\( (read-char port))) (error "coding error"))
 	  (if (char=? (peek-char port) #\() (read-char port))
 	  (let ((word (read-word cs:rparen)))
-	    ;;(sf "index word=~S\n" word)
+	    (db "index word=~S\n" word)
 	    (if (not (char=? #\) (read-char port))) (error "coding error"))
 	    `(index . ,(foldin word '())))))
        
@@ -250,29 +257,37 @@
        (read-vref
 	(lambda ()
 	  (let* ((frag (read-frag cs:vt))
-		 (frag (sx-ref frag 1)) ; extract string value
+		 (iden (sx-ref frag 1)) ; extract string value
 		 (ch1 (peek-char port))
 		 (indx (cond ((eof-object? ch1) #f)
 			     ((char=? ch1 #\() #t)
-			     (else #f))))
-	    (db "$frag=~S ch1=~S\n" frag ch1)
-	    (if indx
-		      `(deref-indexed ,frag ,(cadr (read-index)))
-		      `(deref ,frag)))))
+			     (else #f)))
+                 (res (if indx
+		          `(deref-indexed ,iden ,(cadr (read-index)))
+		          `(deref ,iden))))
+	    (db "$frag ~S ch1=~S\n" iden ch1)
+            (set-source-properties! res (source-properties frag))
+            res)))
 
        (init-blev			; initial brace level
 	(lambda (end-cs) (if (eq? end-cs cs:rcurly) 1 0)))
-       
+
        (read-frag
+        (lambda (end-cs)
+          (let* ((srcl (port-line port))
+                 (frag (read-frag-1 end-cs)))
+            (unless (eq? end-cs frag)
+              (set-source-properties! frag `((line . ,srcl) (filename . ,srcf))))
+            frag)))
+        
+       (read-frag-1
 	(lambda (end-cs)
-	  (db "\n== read frag ~S\n" end-cs)
+	  (db "\n== read frag, until ~S\n" end-cs)
 	  (let loop ((tag #f)		     ; tag: string etc
 		     (chl '())		     ; list of chars read
 		     (bl (init-blev end-cs)) ; brace level
 		     (ch (read-char port))) ; next char
 	    (db "F:     tag=~S ch=~S chl=~S bl=~S\n" tag ch chl bl)
-	    ;;(db "F:     tag=~S ch=~S chl=~S bl=~S end-cs=~S\n"
-		;;tag ch chl bl end-cs)
 	    (cond
 	     ((eof-object? ch)
 	      (if (positive? bl) (error "missing end-brace"))
@@ -289,15 +304,15 @@
  	      (if tag (finish tag chl ch) (read-vref)))
 	     ((char=? ch #\{)
 	      (read-brace))
-	     ((char=? ch #\()
+	     #;((char=? ch #\()
 	      (if tag (finish tag chl ch) (read-index)))
 	     ((char=? ch #\[)
-	      (if tag (finish tag chl ch))
-	      (swallow (read-cmmd cs:rs)))
+	      (if tag (finish tag chl ch)
+	          (swallow (read-cmmd cs:rs)))
+              )
 	     ((char=? ch #\")
 	      ;;(swallow (read-frag cs:dquote))
-	      (loop 'string chl bl (read-char port))
-	      )
+	      (loop 'string chl bl (read-char port)))
 	     (else
 	      (loop (or tag 'string) (cons ch chl) bl (read-char port)))))))
        )
@@ -309,6 +324,8 @@
       (cond
        ((eof-object? nxt))
        ((char-set-contains? cs:ct nxt) (read-char port)))
+      (unless (eof-object? cmd)
+        (set-source-properties! cmd (source-properties (cadr cmd))))
       cmd)
     ;; or (swallow (read-cmmd cs:ct))
     ))
@@ -324,6 +341,7 @@
    (call-with-input-string body
      (lambda (port)
        (let loop ((cmd (read-command port)))
+         ;;(pp cmd)
 	 (cond
 	  ((eof-object? cmd) '())
 	  ((null? (cdr cmd)) (loop (read-command port)))
@@ -331,7 +349,8 @@
 	  (else (cons cmd (loop (read-command port))))))))))
 
 (define (fix-expr-string xstr)
-  (let* ((cmmd (call-with-input-string (string-append "expr " xstr)
+  (let* ((xstr (string-trim xstr char-set:whitespace))
+         (cmmd (call-with-input-string (string-append "expr " xstr)
 		(lambda (port) (read-command port))))
 	 (tail (sx-tail cmmd 2))
 	 (tail (splice-xtail tail)))
@@ -354,7 +373,6 @@
 			  (cons term toks)
 			  (cons term toks))))
 		'() (cdr terms))))
-    ;;(sf "sxt ~S => ~S\n" tail toks)
     toks))
 
 ;; @deffn {Procedure} cnvt-args astr
@@ -419,7 +437,9 @@
 (define (num-string cstr)
   (and=>
    (with-input-from-string cstr
-     (lambda () (read-basic-num (read-char) #:signed #t)))
+     (lambda ()
+       (let ((val (read-basic-num (read-char) #:signed #t)))
+	 (and (eof-object? (read-char)) val))))
    (lambda (pair) `(number ,(cdr pair)))))
 
 ;; @deffn {Procedure} vec-string str
@@ -445,14 +465,12 @@
        (string ,body-part) . ,rest)
      (cons
       `(elseif ,(fix-expr-string cnd) ,(split-body body-part))
-      (cnvt-cond-tail (list-tail ctail 3))))
+      (cnvt-cond-tail (list-tail ctail 4))))
     (`((string "elseif") (string ,cnd) (string ,body-part) . ,rest)
      (cons
       `(elseif ,(fix-expr-string cnd) ,(split-body body-part))
       (cnvt-cond-tail (list-tail ctail 3))))
     (_
-     ;;(sf "ctail: @ ~S\n" (port-line (current-input-port)))
-     ;;(pp ctail) (quit)
      (throw 'tcl-error "syntax error"))))
     
 
@@ -484,11 +502,9 @@
      . ,(lambda (tree)
 	  ;;(sf "tree:\n") (pp tree) ;;(quit)
 	  ;;(sf "============\n")
-	  ;;(pp
 	  (sxml-match tree
 	    ((command (string "if") (string ,cnd) (string "then")
 		      (string ,bdy) . ,rest)
-	     ;;(sf "GOT IT\n") (quit)
 	     `(if ,(fix-expr-string cnd) ,(split-body bdy)
 		  . ,(cnvt-cond-tail rest)))
 	    ((command (string "if") (string ,cnd) (string ,bdy) . ,rest)
@@ -506,10 +522,13 @@
 	    ((command (string "proc") (string ,name) (string ,args)
 		      (string ,body))
 	     `(proc ,name ,(cnvt-args args) ,(split-body body)))
-	    (,_ (report-error "usage: proc name args body")))))
+            ;;((command (string "proc") (word x) (string ,args) (string ,body))
+	    ;;(,_ (report-error "usage: proc name args body"))
+            (,_ #f)
+            )))
     ("return"
      . ,(lambda (tree)
-	  `(return ,(sx-ref tree 2))))
+	  `(return ,(or (sx-ref tree 2) ""))))
     ("set"
      . ,(lambda (tree)
 	  (sxml-match tree
@@ -560,7 +579,6 @@
 ;; (cnvt-tcl '(command (string "expr") ...)) => (expr ...)
 ;; @end example
 (define (cnvt-tree tree)
-  ;;(sf "tree=~S\n" tree)
   (letrec
       ((cnvt-elt
 	(lambda (tree)
@@ -583,16 +601,12 @@
 		 (let* ((tag (sx-tag tree))
 			(tail0 (sx-tail tree))
 			(tail1 (cnvt-tail tail0)))
-		   ;;(sf "cnvt _ tag=~S\n" tag)
-		   ;;(sf "     tail0=~S tail1=~S\n" tail0 tail1)
 		   (if (eq? tail1 tail0) tree (cons tag tail1))))))))
        (cnvt-tail
 	(lambda (tail)
 	  (if (null? tail) tail
 	      (let* ((head0 (car tail)) (head1 (cnvt-elt head0))
 		     (tail0 (cdr tail)) (tail1 (cnvt-tail tail0)))
-		;;(sf "cnvt-tail ~S ~S\n" head0 tail0)
-		;;(sf "cnvt-tail ~S ~S\n" head1 tail1)
 		(if (eq? head1 head0)
 		    (if (eq? tail1 tail0)
 			tail
@@ -610,8 +624,6 @@
      Guile extension language routine to read a single statement."
   (let* ((cmmd0 (read-command port))
 	 (cmmd1 (cnvt-tree cmmd0)))
-    ;;(sf "s:cmd1:\n ") (pp cmmd0)
-    ;;(sf "s:cmd2:\n ") (pp cmmd1)
     cmmd1))
 
 ;; @deffn {Procedure} read-tcl-file port env
@@ -627,10 +639,6 @@
        (let loop ((cmd (read-command port)))
 	 (if (eof-object? cmd) '()
 	     (let ((cmd1 cmd) (cmd2 (cnvt-tree cmd)))
-	       ;;(sf "cmd1:\n") (pp cmd1)
-	       ;;(sf "cmd2::n") (pp cmd2)
-	       (cons (cnvt-tree cmd) (loop (read-command port))))))))
-  )
-
+	       (cons (cnvt-tree cmd) (loop (read-command port)))))))))
 
 ;; --- last line ---
